{ANTOAN_NOP,"ANTOAN_NOP"},
{ANTOAN_UNKNOWN_COMMAND,"ANTOAN_UNKNOWN_COMMAND"},
{ANTOAN_BAD_COMMAND_SIZE,"ANTOAN_BAD_COMMAND_SIZE"},
{ANTOAN_FORCE_TIMEOUT,"ANTOAN_FORCE_TIMEOUT"},
{ANTOAN_GET_VERSION,"ANTOAN_GET_VERSION"},
{ANTOAN_VERSION,"ANTOAN_VERSION"},
{ANTOMA_REGISTER,"ANTOMA_REGISTER"},
{MATOAN_METACHANGES_LOG,"MATOAN_METACHANGES_LOG"},
{MATOAN_MASTER_ACK,"MATOAN_MASTER_ACK"},
{MATOAN_STATE,"MATOAN_STATE"},
{ANTOMA_FORCE_METADATA_SAVE,"ANTOMA_FORCE_METADATA_SAVE"},
{MATOMA_METADATA_CHECKSUM,"MATOMA_METADATA_CHECKSUM"},
{MATOMA_FORCE_DESYNC,"MATOMA_FORCE_DESYNC"},
{ANTOMA_DOWNLOAD_START,"ANTOMA_DOWNLOAD_START"},
{MATOAN_DOWNLOAD_INFO,"MATOAN_DOWNLOAD_INFO"},
{ANTOMA_DOWNLOAD_REQUEST,"ANTOMA_DOWNLOAD_REQUEST"},
{MATOAN_DOWNLOAD_DATA,"MATOAN_DOWNLOAD_DATA"},
{ANTOMA_DOWNLOAD_END,"ANTOMA_DOWNLOAD_END"},
{ANTOMA_STORE_METADATA,"ANTOMA_STORE_METADATA"},
{ANTOMA_SYSLOG,"ANTOMA_SYSLOG"},
{ANTOAN_GET_CONFIG,"ANTOAN_GET_CONFIG"},
{ANTOAN_CONFIG_VALUE,"ANTOAN_CONFIG_VALUE"},
{ANTOAN_GET_CONFIG_FILE,"ANTOAN_GET_CONFIG_FILE"},
{ANTOAN_CONFIG_FILE_CONTENT,"ANTOAN_CONFIG_FILE_CONTENT"},
{ANTOAN_GET_CONFIG_FILE_MD5,"ANTOAN_GET_CONFIG_FILE_MD5"},
{ANTOAN_CONFIG_FILE_MD5,"ANTOAN_CONFIG_FILE_MD5"},
{CSTOMA_CHUNK_DOESNT_EXIST,"CSTOMA_CHUNK_DOESNT_EXIST"},
{MATOCS_CHUNK_STATUS,"MATOCS_CHUNK_STATUS"},
{CSTOMA_CHUNK_STATUS,"CSTOMA_CHUNK_STATUS"},
{MATOCS_REGISTER_FIRST,"MATOCS_REGISTER_FIRST"},
{CSTOMA_REGISTER,"CSTOMA_REGISTER"},
{CSTOMA_SPACE,"CSTOMA_SPACE"},
{CSTOMA_CHUNK_DAMAGED,"CSTOMA_CHUNK_DAMAGED"},
{CSTOMA_CURRENT_LOAD,"CSTOMA_CURRENT_LOAD"},
{MATOCS_MASTER_ACK,"MATOCS_MASTER_ACK"},
{CSTOMA_CHUNK_LOST,"CSTOMA_CHUNK_LOST"},
{CSTOMA_ERROR_OCCURRED,"CSTOMA_ERROR_OCCURRED"},
{CSTOMA_CHUNK_NEW,"CSTOMA_CHUNK_NEW"},
{CSTOMA_LABELS,"CSTOMA_LABELS"},
{MATOCS_CREATE,"MATOCS_CREATE"},
{CSTOMA_CREATE,"CSTOMA_CREATE"},
{MATOCS_DELETE,"MATOCS_DELETE"},
{CSTOMA_DELETE,"CSTOMA_DELETE"},
{MATOCS_DUPLICATE,"MATOCS_DUPLICATE"},
{CSTOMA_DUPLICATE,"CSTOMA_DUPLICATE"},
{MATOCS_SET_VERSION,"MATOCS_SET_VERSION"},
{CSTOMA_SET_VERSION,"CSTOMA_SET_VERSION"},
{MATOCS_REPLICATE,"MATOCS_REPLICATE"},
{CSTOMA_REPLICATE,"CSTOMA_REPLICATE"},
{MATOCS_CHUNKOP,"MATOCS_CHUNKOP"},
{CSTOMA_CHUNKOP,"CSTOMA_CHUNKOP"},
{MATOCS_REPLICATE_SPLIT,"MATOCS_REPLICATE_SPLIT"},
{CSTOMA_REPLICATE_SPLIT,"CSTOMA_REPLICATE_SPLIT"},
{MATOCS_REPLICATE_RECOVER,"MATOCS_REPLICATE_RECOVER"},
{CSTOMA_REPLICATE_RECOVER,"CSTOMA_REPLICATE_RECOVER"},
{MATOCS_REPLICATE_JOIN,"MATOCS_REPLICATE_JOIN"},
{CSTOMA_REPLICATE_JOIN,"CSTOMA_REPLICATE_JOIN"},
{MATOCS_TRUNCATE,"MATOCS_TRUNCATE"},
{CSTOMA_TRUNCATE,"CSTOMA_TRUNCATE"},
{MATOCS_DUPTRUNC,"MATOCS_DUPTRUNC"},
{CSTOMA_DUPTRUNC,"CSTOMA_DUPTRUNC"},
{MATOCS_LOCALSPLIT,"MATOCS_LOCALSPLIT"},
{CSTOMA_LOCALSPLIT,"CSTOMA_LOCALSPLIT"},
{CLTOCS_READ,"CLTOCS_READ"},
{CSTOCL_READ_STATUS,"CSTOCL_READ_STATUS"},
{CSTOCL_READ_DATA,"CSTOCL_READ_DATA"},
{CLTOCS_WRITE,"CLTOCS_WRITE"},
{CSTOCL_WRITE_STATUS,"CSTOCL_WRITE_STATUS"},
{CLTOCS_WRITE_DATA,"CLTOCS_WRITE_DATA"},
{CLTOCS_WRITE_FINISH,"CLTOCS_WRITE_FINISH"},
{ANTOCS_GET_CHUNK_BLOCKS,"ANTOCS_GET_CHUNK_BLOCKS"},
{CSTOAN_CHUNK_BLOCKS,"CSTOAN_CHUNK_BLOCKS"},
{ANTOCS_GET_CHUNK_CHECKSUM,"ANTOCS_GET_CHUNK_CHECKSUM"},
{CSTOAN_CHUNK_CHECKSUM,"CSTOAN_CHUNK_CHECKSUM"},
{ANTOCS_GET_CHUNK_CHECKSUM_TAB,"ANTOCS_GET_CHUNK_CHECKSUM_TAB"},
{CSTOAN_CHUNK_CHECKSUM_TAB,"CSTOAN_CHUNK_CHECKSUM_TAB"},
{ANTOCS_GET_CHUNK_INFO,"ANTOCS_GET_CHUNK_INFO"},
{CSTOAN_CHUNK_INFO,"CSTOAN_CHUNK_INFO"},
{ANTOCS_CLEAR_ERRORS,"ANTOCS_CLEAR_ERRORS"},
{CSTOAN_CLEAR_ERRORS,"CSTOAN_CLEAR_ERRORS"},
{CLTOMA_SCLASS_CREATE,"CLTOMA_SCLASS_CREATE"},
{MATOCL_SCLASS_CREATE,"MATOCL_SCLASS_CREATE"},
{CLTOMA_SCLASS_CHANGE,"CLTOMA_SCLASS_CHANGE"},
{MATOCL_SCLASS_CHANGE,"MATOCL_SCLASS_CHANGE"},
{CLTOMA_SCLASS_DELETE,"CLTOMA_SCLASS_DELETE"},
{MATOCL_SCLASS_DELETE,"MATOCL_SCLASS_DELETE"},
{CLTOMA_SCLASS_DUPLICATE,"CLTOMA_SCLASS_DUPLICATE"},
{MATOCL_SCLASS_DUPLICATE,"MATOCL_SCLASS_DUPLICATE"},
{CLTOMA_SCLASS_RENAME,"CLTOMA_SCLASS_RENAME"},
{MATOCL_SCLASS_RENAME,"MATOCL_SCLASS_RENAME"},
{CLTOMA_SCLASS_LIST,"CLTOMA_SCLASS_LIST"},
{MATOCL_SCLASS_LIST,"MATOCL_SCLASS_LIST"},
{CLTOMA_PATTERN_ADD,"CLTOMA_PATTERN_ADD"},
{MATOCL_PATTERN_ADD,"MATOCL_PATTERN_ADD"},
{CLTOMA_PATTERN_DELETE,"CLTOMA_PATTERN_DELETE"},
{MATOCL_PATTERN_DELETE,"MATOCL_PATTERN_DELETE"},
{CLTOMA_PATTERN_LIST,"CLTOMA_PATTERN_LIST"},
{MATOCL_PATTERN_LIST,"MATOCL_PATTERN_LIST"},
{CLTOMA_TRASH_LIST,"CLTOMA_TRASH_LIST"},
{MATOCL_TRASH_LIST,"MATOCL_TRASH_LIST"},
{CLTOMA_TRASH_RECOVER,"CLTOMA_TRASH_RECOVER"},
{MATOCL_TRASH_RECOVER,"MATOCL_TRASH_RECOVER"},
{CLTOMA_TRASH_REMOVE,"CLTOMA_TRASH_REMOVE"},
{MATOCL_TRASH_REMOVE,"MATOCL_TRASH_REMOVE"},
{CLTOMA_SUSTAINED_LIST,"CLTOMA_SUSTAINED_LIST"},
{MATOCL_SUSTAINED_LIST,"MATOCL_SUSTAINED_LIST"},
{CLTOMA_PATH_LOOKUP,"CLTOMA_PATH_LOOKUP"},
{MATOCL_PATH_LOOKUP,"MATOCL_PATH_LOOKUP"},
{CLTOMA_FUSE_REGISTER,"CLTOMA_FUSE_REGISTER"},
{MATOCL_FUSE_REGISTER,"MATOCL_FUSE_REGISTER"},
{CLTOMA_FUSE_STATFS,"CLTOMA_FUSE_STATFS"},
{MATOCL_FUSE_STATFS,"MATOCL_FUSE_STATFS"},
{CLTOMA_FUSE_ACCESS,"CLTOMA_FUSE_ACCESS"},
{MATOCL_FUSE_ACCESS,"MATOCL_FUSE_ACCESS"},
{CLTOMA_FUSE_LOOKUP,"CLTOMA_FUSE_LOOKUP"},
{MATOCL_FUSE_LOOKUP,"MATOCL_FUSE_LOOKUP"},
{CLTOMA_FUSE_GETATTR,"CLTOMA_FUSE_GETATTR"},
{MATOCL_FUSE_GETATTR,"MATOCL_FUSE_GETATTR"},
{CLTOMA_FUSE_SETATTR,"CLTOMA_FUSE_SETATTR"},
{MATOCL_FUSE_SETATTR,"MATOCL_FUSE_SETATTR"},
{CLTOMA_FUSE_READLINK,"CLTOMA_FUSE_READLINK"},
{MATOCL_FUSE_READLINK,"MATOCL_FUSE_READLINK"},
{CLTOMA_FUSE_SYMLINK,"CLTOMA_FUSE_SYMLINK"},
{MATOCL_FUSE_SYMLINK,"MATOCL_FUSE_SYMLINK"},
{CLTOMA_FUSE_MKNOD,"CLTOMA_FUSE_MKNOD"},
{MATOCL_FUSE_MKNOD,"MATOCL_FUSE_MKNOD"},
{CLTOMA_FUSE_MKDIR,"CLTOMA_FUSE_MKDIR"},
{MATOCL_FUSE_MKDIR,"MATOCL_FUSE_MKDIR"},
{CLTOMA_FUSE_UNLINK,"CLTOMA_FUSE_UNLINK"},
{MATOCL_FUSE_UNLINK,"MATOCL_FUSE_UNLINK"},
{CLTOMA_FUSE_RMDIR,"CLTOMA_FUSE_RMDIR"},
{MATOCL_FUSE_RMDIR,"MATOCL_FUSE_RMDIR"},
{CLTOMA_FUSE_RENAME,"CLTOMA_FUSE_RENAME"},
{MATOCL_FUSE_RENAME,"MATOCL_FUSE_RENAME"},
{CLTOMA_FUSE_LINK,"CLTOMA_FUSE_LINK"},
{MATOCL_FUSE_LINK,"MATOCL_FUSE_LINK"},
{CLTOMA_FUSE_READDIR,"CLTOMA_FUSE_READDIR"},
{MATOCL_FUSE_READDIR,"MATOCL_FUSE_READDIR"},
{CLTOMA_FUSE_OPEN,"CLTOMA_FUSE_OPEN"},
{MATOCL_FUSE_OPEN,"MATOCL_FUSE_OPEN"},
{CLTOMA_FUSE_READ_CHUNK,"CLTOMA_FUSE_READ_CHUNK"},
{MATOCL_FUSE_READ_CHUNK,"MATOCL_FUSE_READ_CHUNK"},
{CLTOMA_FUSE_WRITE_CHUNK,"CLTOMA_FUSE_WRITE_CHUNK"},
{MATOCL_FUSE_WRITE_CHUNK,"MATOCL_FUSE_WRITE_CHUNK"},
{CLTOMA_FUSE_WRITE_CHUNK_END,"CLTOMA_FUSE_WRITE_CHUNK_END"},
{MATOCL_FUSE_WRITE_CHUNK_END,"MATOCL_FUSE_WRITE_CHUNK_END"},
{CLTOMA_FUSE_APPEND_SLICE,"CLTOMA_FUSE_APPEND_SLICE"},
{MATOCL_FUSE_APPEND_SLICE,"MATOCL_FUSE_APPEND_SLICE"},
{CLTOMA_FUSE_CHECK,"CLTOMA_FUSE_CHECK"},
{MATOCL_FUSE_CHECK,"MATOCL_FUSE_CHECK"},
{CLTOMA_FUSE_GETTRASHRETENTION,"CLTOMA_FUSE_GETTRASHRETENTION"},
{MATOCL_FUSE_GETTRASHRETENTION,"MATOCL_FUSE_GETTRASHRETENTION"},
{CLTOMA_FUSE_SETTRASHRETENTION,"CLTOMA_FUSE_SETTRASHRETENTION"},
{MATOCL_FUSE_SETTRASHRETENTION,"MATOCL_FUSE_SETTRASHRETENTION"},
{CLTOMA_FUSE_GETSCLASS,"CLTOMA_FUSE_GETSCLASS"},
{MATOCL_FUSE_GETSCLASS,"MATOCL_FUSE_GETSCLASS"},
{CLTOMA_FUSE_SETSCLASS,"CLTOMA_FUSE_SETSCLASS"},
{MATOCL_FUSE_SETSCLASS,"MATOCL_FUSE_SETSCLASS"},
{CLTOMA_FUSE_GETTRASH,"CLTOMA_FUSE_GETTRASH"},
{MATOCL_FUSE_GETTRASH,"MATOCL_FUSE_GETTRASH"},
{CLTOMA_FUSE_GETDETACHEDATTR,"CLTOMA_FUSE_GETDETACHEDATTR"},
{MATOCL_FUSE_GETDETACHEDATTR,"MATOCL_FUSE_GETDETACHEDATTR"},
{CLTOMA_FUSE_GETTRASHPATH,"CLTOMA_FUSE_GETTRASHPATH"},
{MATOCL_FUSE_GETTRASHPATH,"MATOCL_FUSE_GETTRASHPATH"},
{CLTOMA_FUSE_SETTRASHPATH,"CLTOMA_FUSE_SETTRASHPATH"},
{MATOCL_FUSE_SETTRASHPATH,"MATOCL_FUSE_SETTRASHPATH"},
{CLTOMA_FUSE_UNDEL,"CLTOMA_FUSE_UNDEL"},
{MATOCL_FUSE_UNDEL,"MATOCL_FUSE_UNDEL"},
{CLTOMA_FUSE_PURGE,"CLTOMA_FUSE_PURGE"},
{MATOCL_FUSE_PURGE,"MATOCL_FUSE_PURGE"},
{CLTOMA_FUSE_GETDIRSTATS,"CLTOMA_FUSE_GETDIRSTATS"},
{MATOCL_FUSE_GETDIRSTATS,"MATOCL_FUSE_GETDIRSTATS"},
{CLTOMA_FUSE_TRUNCATE,"CLTOMA_FUSE_TRUNCATE"},
{MATOCL_FUSE_TRUNCATE,"MATOCL_FUSE_TRUNCATE"},
{CLTOMA_FUSE_REPAIR,"CLTOMA_FUSE_REPAIR"},
{MATOCL_FUSE_REPAIR,"MATOCL_FUSE_REPAIR"},
{CLTOMA_FUSE_SNAPSHOT,"CLTOMA_FUSE_SNAPSHOT"},
{MATOCL_FUSE_SNAPSHOT,"MATOCL_FUSE_SNAPSHOT"},
{CLTOMA_FUSE_GETSUSTAINED,"CLTOMA_FUSE_GETSUSTAINED"},
{MATOCL_FUSE_GETSUSTAINED,"MATOCL_FUSE_GETSUSTAINED"},
{CLTOMA_FUSE_GETEATTR,"CLTOMA_FUSE_GETEATTR"},
{MATOCL_FUSE_GETEATTR,"MATOCL_FUSE_GETEATTR"},
{CLTOMA_FUSE_SETEATTR,"CLTOMA_FUSE_SETEATTR"},
{MATOCL_FUSE_SETEATTR,"MATOCL_FUSE_SETEATTR"},
{CLTOMA_FUSE_QUOTACONTROL,"CLTOMA_FUSE_QUOTACONTROL"},
{MATOCL_FUSE_QUOTACONTROL,"MATOCL_FUSE_QUOTACONTROL"},
{CLTOMA_FUSE_GETXATTR,"CLTOMA_FUSE_GETXATTR"},
{MATOCL_FUSE_GETXATTR,"MATOCL_FUSE_GETXATTR"},
{CLTOMA_FUSE_SETXATTR,"CLTOMA_FUSE_SETXATTR"},
{MATOCL_FUSE_SETXATTR,"MATOCL_FUSE_SETXATTR"},
{CLTOMA_FUSE_CREATE,"CLTOMA_FUSE_CREATE"},
{MATOCL_FUSE_CREATE,"MATOCL_FUSE_CREATE"},
{CLTOMA_FUSE_PARENTS,"CLTOMA_FUSE_PARENTS"},
{MATOCL_FUSE_PARENTS,"MATOCL_FUSE_PARENTS"},
{CLTOMA_FUSE_PATHS,"CLTOMA_FUSE_PATHS"},
{MATOCL_FUSE_PATHS,"MATOCL_FUSE_PATHS"},
{CLTOMA_FUSE_GETFACL,"CLTOMA_FUSE_GETFACL"},
{MATOCL_FUSE_GETFACL,"MATOCL_FUSE_GETFACL"},
{CLTOMA_FUSE_SETFACL,"CLTOMA_FUSE_SETFACL"},
{MATOCL_FUSE_SETFACL,"MATOCL_FUSE_SETFACL"},
{CLTOMA_FUSE_FLOCK,"CLTOMA_FUSE_FLOCK"},
{MATOCL_FUSE_FLOCK,"MATOCL_FUSE_FLOCK"},
{CLTOMA_FUSE_POSIX_LOCK,"CLTOMA_FUSE_POSIX_LOCK"},
{MATOCL_FUSE_POSIX_LOCK,"MATOCL_FUSE_POSIX_LOCK"},
{CLTOMA_FUSE_ARCHCTL,"CLTOMA_FUSE_ARCHCTL"},
{MATOCL_FUSE_ARCHCTL,"MATOCL_FUSE_ARCHCTL"},
{CLTOMA_FUSE_FSYNC,"CLTOMA_FUSE_FSYNC"},
{MATOCL_FUSE_FSYNC,"MATOCL_FUSE_FSYNC"},
{CLTOMA_FUSE_SUSTAINED_INODES_DEPRECATED,"CLTOMA_FUSE_SUSTAINED_INODES_DEPRECATED"},
{CLTOMA_CSERV_LIST,"CLTOMA_CSERV_LIST"},
{MATOCL_CSERV_LIST,"MATOCL_CSERV_LIST"},
{CLTOAN_MONOTONIC_DATA,"CLTOAN_MONOTONIC_DATA"},
{ANTOCL_MONOTONIC_DATA,"ANTOCL_MONOTONIC_DATA"},
{CLTOAN_CHART,"CLTOAN_CHART"},
{ANTOCL_CHART,"ANTOCL_CHART"},
{CLTOAN_CHART_DATA,"CLTOAN_CHART_DATA"},
{ANTOCL_CHART_DATA,"ANTOCL_CHART_DATA"},
{CLTOMA_SESSION_LIST,"CLTOMA_SESSION_LIST"},
{MATOCL_SESSION_LIST,"MATOCL_SESSION_LIST"},
{CLTOMA_INFO,"CLTOMA_INFO"},
{MATOCL_INFO,"MATOCL_INFO"},
{CLTOMA_FSTEST_INFO,"CLTOMA_FSTEST_INFO"},
{MATOCL_FSTEST_INFO,"MATOCL_FSTEST_INFO"},
{CLTOMA_CHUNKSTEST_INFO,"CLTOMA_CHUNKSTEST_INFO"},
{MATOCL_CHUNKSTEST_INFO,"MATOCL_CHUNKSTEST_INFO"},
{CLTOMA_CHUNKS_MATRIX,"CLTOMA_CHUNKS_MATRIX"},
{MATOCL_CHUNKS_MATRIX,"MATOCL_CHUNKS_MATRIX"},
{CLTOMA_QUOTA_INFO,"CLTOMA_QUOTA_INFO"},
{MATOCL_QUOTA_INFO,"MATOCL_QUOTA_INFO"},
{CLTOMA_EXPORTS_INFO,"CLTOMA_EXPORTS_INFO"},
{MATOCL_EXPORTS_INFO,"MATOCL_EXPORTS_INFO"},
{CLTOMA_MLOG_LIST,"CLTOMA_MLOG_LIST"},
{MATOCL_MLOG_LIST,"MATOCL_MLOG_LIST"},
{CLTOMA_CSSERV_COMMAND,"CLTOMA_CSSERV_COMMAND"},
{MATOCL_CSSERV_COMMAND,"MATOCL_CSSERV_COMMAND"},
{CLTOMA_SESSION_COMMAND,"CLTOMA_SESSION_COMMAND"},
{MATOCL_SESSION_COMMAND,"MATOCL_SESSION_COMMAND"},
{CLTOMA_MEMORY_INFO,"CLTOMA_MEMORY_INFO"},
{MATOCL_MEMORY_INFO,"MATOCL_MEMORY_INFO"},
{CLTOAN_MODULE_INFO,"CLTOAN_MODULE_INFO"},
{ANTOCL_MODULE_INFO,"ANTOCL_MODULE_INFO"},
{CLTOMA_LIST_OPEN_FILES,"CLTOMA_LIST_OPEN_FILES"},
{MATOCL_LIST_OPEN_FILES,"MATOCL_LIST_OPEN_FILES"},
{CLTOMA_LIST_ACQUIRED_LOCKS,"CLTOMA_LIST_ACQUIRED_LOCKS"},
{MATOCL_LIST_ACQUIRED_LOCKS,"MATOCL_LIST_ACQUIRED_LOCKS"},
{CLTOMA_MASS_RESOLVE_PATHS,"CLTOMA_MASS_RESOLVE_PATHS"},
{MATOCL_MASS_RESOLVE_PATHS,"MATOCL_MASS_RESOLVE_PATHS"},
{CLTOMA_SCLASS_INFO,"CLTOMA_SCLASS_INFO"},
{MATOCL_SCLASS_INFO,"MATOCL_SCLASS_INFO"},
{CLTOMA_MISSING_CHUNKS,"CLTOMA_MISSING_CHUNKS"},
{MATOCL_MISSING_CHUNKS,"MATOCL_MISSING_CHUNKS"},
{CLTOMA_NODE_INFO,"CLTOMA_NODE_INFO"},
{MATOCL_NODE_INFO,"MATOCL_NODE_INFO"},
{CLTOMA_PATTERN_INFO,"CLTOMA_PATTERN_INFO"},
{MATOCL_PATTERN_INFO,"MATOCL_PATTERN_INFO"},
{CLTOMA_INSTANCE_NAME,"CLTOMA_INSTANCE_NAME"},
{MATOCL_INSTANCE_NAME,"MATOCL_INSTANCE_NAME"},
{CLTOMA_FULL_DIRECTORY_DATA,"CLTOMA_FULL_DIRECTORY_DATA"},
{MATOCL_FULL_DIRECTORY_DATA,"MATOCL_FULL_DIRECTORY_DATA"},
{CLTOMA_SET_ALL_NODE_ATTRIBUTES,"CLTOMA_SET_ALL_NODE_ATTRIBUTES"},
{MATOCL_SET_ALL_NODE_ATTRIBUTES,"MATOCL_SET_ALL_NODE_ATTRIBUTES"},
{CLTOCS_HDD_LIST,"CLTOCS_HDD_LIST"},
{CSTOCL_HDD_LIST,"CSTOCL_HDD_LIST"},
{CLTOMA_FUSE_SUSTAINED_INODES,"CLTOMA_FUSE_SUSTAINED_INODES"},
{CLTOMA_FUSE_AMTIME_INODES,"CLTOMA_FUSE_AMTIME_INODES"},
{MATOCL_FUSE_CHUNK_HAS_CHANGED,"MATOCL_FUSE_CHUNK_HAS_CHANGED"},
{MATOCL_FUSE_FLENG_HAS_CHANGED,"MATOCL_FUSE_FLENG_HAS_CHANGED"},
{CLTOMA_FUSE_TIME_SYNC,"CLTOMA_FUSE_TIME_SYNC"},
{MATOCL_FUSE_TIME_SYNC,"MATOCL_FUSE_TIME_SYNC"},
{MATOCL_FUSE_INVALIDATE_CHUNK_CACHE,"MATOCL_FUSE_INVALIDATE_CHUNK_CACHE"},
{CLTOMA_FUSE_OPDATA,"CLTOMA_FUSE_OPDATA"},
{CLTOMA_FUSE_WFLAGS,"CLTOMA_FUSE_WFLAGS"},
